#include <stdlib.h>
#include <stdio.h>
#include "list.h"

// segnature delle funzioni ausiliarie (private a questo file,
// quindi non inserite in list.h)
void toString_aux(char *res, struct list *this);
struct list *copyOf(struct list *this);

struct list *construct(int head, struct list *tail) {
  struct list *this = malloc(sizeof(struct list));

  this->head = head;
  this->tail = tail;

  return this;
}


// Definiamo una funzione toString() che, in generale, traduce in stringhe 
// liste di interi anche negativi.
char *toString(struct list *this) {

  // Allochiamo la stringa che rappresenta la lista.
  // Per ogni intrero della lista allochiamo
  // 10 caratteri per le cifre che può avere un intero
  //  1 carattere per il segno
  //  1 un carattere per la virgola che separa i valori
  // allochiamo inoltre altri tre caratteri:
  //  2 caratteri per le parentesi '[' e ']'
  //  1 carattere per il fine stringa '\0'
  int l = length(this);
  char *res = malloc(l * 12 + 3);

  *res = '[';
  toString_aux(res + 1, this);

  return res;
}


void toString_aux(char *res, struct list *this) {
  if (this && this->tail)
    toString_aux(res + sprintf(res, "%d,", this->head), this->tail);
  else if (this)
    sprintf(res, "%d]", this->head);
  else
    sprintf(res, "]");
}


int length(struct list *this) {
  if (!this)
    return 0;
  else
    return 1 + length(this->tail);
}


struct list *read() {
  struct list *temp = NULL;
  struct list *res = NULL;
  int n;

  do {
    printf("? ");
    scanf("%d", &n);
    if (n > 0)
      if (temp)
        temp = temp->tail = construct(n, NULL);
      else 
        temp = res = construct(n, NULL);
  } while (n > 0);

  return res;
}


// Questa funzione restituisce una copia di this.
// Viene usata dalla funzione fusion per restituire una stringa completamente 
// indipendente dalle stringhe this e that passate come parametri, in questo 
// modo si evitano dei possibili side-effects su this o that quando si va a 
// modificare la stringa risultato della funzione fusion.
struct list *copyOf(struct list *this) {
  if (this == NULL)
    return NULL;
  return construct(this->head, copyOf(this->tail));
}


struct list *fusion(struct list *this, struct list *that) {
  if (!this)
    return copyOf(that); // restituiamo una copia per evitare side-effects
  else if (!that) 
    return copyOf(this); // restituiamo una copia per evitare side-effects
  else if (this->head <= that->head)
    return construct(this->head, fusion(this->tail, that));
  else
    return construct(that->head, fusion(this, that->tail));
}
